//
// Copyright (c) Microsoft Corporation.  All rights reserved.
//
//
// Use of this sample source code is subject to the terms of the Microsoft
// license agreement under which you licensed this sample source code. If
// you did not accept the terms of the license agreement, you are not
// authorized to use this sample source code. For the terms of the license,
// please see the license agreement between you and Microsoft or, if applicable,
// see the LICENSE.RTF on your install media or the root of your tools installation.
// THE SAMPLE SOURCE CODE IS PROVIDED "AS IS", WITH NO WARRANTIES OR INDEMNITIES.
//

//
// Copyright (C) 2011, UNITEC Co., LTD. All Rights Reserved.
// THIS SOURCE CODE, AND ITS USE AND DISTRIBUTION, IS SUBJECT TO THE TERMS
// AND CONDITIONS OF THE APPLICABLE LICENSE AGREEMENT
//

//
//  File: tchstreampdd.cpp
//
#pragma warning(push)
#pragma warning(disable: 4127 4201)

// Public
#include <windows.h>
#include <types.h>
#include <nkintr.h>
#include <creg.hxx>
#include <pm.h>

// Platform
#include <ceddk.h>

// Local
#include <mtouch.h>
#include "tchstreampdd.h"

// Debug zones
#ifndef SHIP_BUILD

#undef ZONE_ERROR
#undef ZONE_WARN
#undef ZONE_FUNCTION
#undef ZONE_INFO
#undef ZONE_TIPSTATE

#define ZONE_ERROR          DEBUGZONE(0)
#define ZONE_WARN           DEBUGZONE(1)
#define ZONE_FUNCTION       DEBUGZONE(2)
#define ZONE_INFO           DEBUGZONE(3)
#define ZONE_TIPSTATE       DEBUGZONE(4)

#endif


//TCH PDD DDSI functions
DWORD WINAPI TchPdd_Init(
    LPCTSTR pszActiveKey,
    TCH_MDD_INTERFACE_INFO* pMddIfc,
    TCH_PDD_INTERFACE_INFO* pPddIfc,
    DWORD hMddContext
    );

void WINAPI TchPdd_Deinit(DWORD hPddContext);
void WINAPI TchPdd_PowerUp(DWORD hPddContext);
void WINAPI TchPdd_PowerDown(DWORD hPddContext);
BOOL WINAPI TchPdd_Ioctl(
    DWORD hPddContext,
    DWORD dwCode,
    PBYTE pBufIn,
    DWORD dwLenIn,
    PBYTE pBufOut,
    DWORD dwLenOut,
    PDWORD pdwActualOut
    );

// globals
static DWORD                          s_mddContext;
static PFN_TCH_MDD_REPORTSAMPLESET    s_pfnMddReportSampleSet;
static HANDLE                         s_hISTThread;

//==============================================================================
// Function Name: TchPdd_Init
//
// Description: PDD should always implement this function. MDD calls it during
//              initialization to fill up the function table with rest of the
//              Helper functions.
//
// Arguments:
//              [IN] pszActiveKey - current active touch driver key
//              [IN] pMddIfc - MDD interface info
//              [OUT]pPddIfc - PDD interface (the function table to be filled)
//              [IN] hMddContext - mdd context (send to MDD in callback fn)
//
// Ret Value:   pddContext.
//==============================================================================
extern "C" DWORD WINAPI TchPdd_Init(LPCTSTR pszActiveKey,
        TCH_MDD_INTERFACE_INFO* pMddIfc,
        TCH_PDD_INTERFACE_INFO* pPddIfc,
        DWORD hMddContext
        )
{
    DWORD pddContext = 0;

    DEBUGMSG(ZONE_FUNCTION, (TEXT("TchPdd_Init+\r\n")));

    RETAILMSG(1, (_T("%S: Driver Version = %s\r\n"), __FUNCTION__, TCH_DRIVER_VERSION));

    // Initialize once only
    if (g_TouchDevice.bInitialized)
    {
        pddContext = (DWORD) &g_TouchDevice;
        goto cleanup;
    }


    // Remember the callback function pointer
    s_pfnMddReportSampleSet = pMddIfc->pfnMddReportSampleSet;

    // Remember the mdd context
    s_mddContext = hMddContext;

    g_TouchDevice.nSampleRate = DEFAULT_SAMPLE_RATE_HZ;
    g_TouchDevice.dwPowerState = D0;
    g_TouchDevice.dwSamplingTimeOut = INFINITE;
    g_TouchDevice.bTerminateIST = FALSE;
    g_TouchDevice.hTouchPanelEvent = 0;

    // Initialize HW
    if (!PDDInitializeHardware(pszActiveKey))
    {
        DEBUGMSG(ZONE_ERROR,  (TEXT("ERROR: TOUCH: Failed to initialize touch PDD.\r\n")));
        goto cleanup;
    }

    //Calibrate the screen, if the calibration data is not already preset in the registry
    PDDStartCalibrationThread();

    //Initialization of the h/w is done
    g_TouchDevice.bInitialized = TRUE;

    pddContext = (DWORD) &g_TouchDevice;

    // fill up pddifc table
    pPddIfc->version        = 1;
    pPddIfc->pfnDeinit      = TchPdd_Deinit;
    pPddIfc->pfnIoctl       = TchPdd_Ioctl;
    pPddIfc->pfnPowerDown   = TchPdd_PowerDown;
    pPddIfc->pfnPowerUp     = TchPdd_PowerUp;

	// Initialize BSP layer
    g_TouchDevice.hTouchPanelEvent = TchBsp_Init(&g_TouchDevice.hBspContext, 
			&g_TouchDevice.BspIfc, &g_TouchDevice.dwMaxContact, g_TouchDevice.dwISTPriority);
	if(!g_TouchDevice.hTouchPanelEvent)
	{
		DEBUGMSG(ZONE_ERROR,(TEXT("PDDTouchPanelEnable: Failed to initialize touch event.\r\n")));
		return NULL;
	}

cleanup:
    DEBUGMSG(ZONE_FUNCTION, (TEXT("TchPdd_Init-\r\n")));
    return pddContext;
}


//==============================================================================
// Function Name: TchPdd_DeInit
//
// Description: MDD calls it during deinitialization. PDD should deinit hardware
//              and deallocate memory etc.
//
// Arguments:   [IN] hPddContext. pddcontext returned in TchPdd_Init
//
// Ret Value:   None
//
//==============================================================================
void WINAPI TchPdd_Deinit(DWORD hPddContext)
{
    DEBUGMSG(ZONE_FUNCTION, (TEXT("TchPdd_Deinit+\r\n")));

    // Remove-W4: Warning C4100 workaround
    UNREFERENCED_PARAMETER(hPddContext);

    // Close the IST and release the resources before
    // de-initializing.
    PDDTouchPanelDisable();

    //  Shutdown HW
    PDDDeinitializeHardware();

    g_TouchDevice.bInitialized = FALSE;


    DEBUGMSG(ZONE_FUNCTION, (TEXT("TchPdd_Deinit-\r\n")));
}


//==============================================================================
// Function Name: TchPdd_Ioctl
//
// Description: The MDD controls the touch PDD through these IOCTLs.
//
// Arguments:   [IN] hPddContext. pddcontext returned in TchPdd_Init
//              DWORD dwCode. IOCTL code
//              PBYTE pBufIn. Input Buffer pointer
//              DWORD dwLenIn. Input buffer length
//              PBYTE pBufOut. Output buffer pointer
//              DWORD dwLenOut. Output buffer length
//              PWORD pdwAcutalOut. Actual output buffer length.
//
// Ret Value:   TRUE if success else FALSE. SetLastError() if FALSE.
//==============================================================================
BOOL WINAPI TchPdd_Ioctl(DWORD hPddContext,
        DWORD dwCode,
        PBYTE pBufIn,
        DWORD dwLenIn,
        PBYTE pBufOut,
        DWORD dwLenOut,
        PDWORD pdwActualOut
        )
{
    DWORD dwResult = ERROR_INVALID_PARAMETER;

    // Remove-W4: Warning C4100 workaround
    UNREFERENCED_PARAMETER(hPddContext);

    // Only allow kernel process to call PDD ioctls.
    if (GetCurrentProcessId() != GetDirectCallerProcessId()) {
        DEBUGMSG(ZONE_ERROR, (L"TchPdd_Ioctl: Can be called only from device process (caller process id 0x%08x)\r\n", \
                    GetDirectCallerProcessId()));
        SetLastError(ERROR_ACCESS_DENIED);
        return FALSE;
    }

    switch (dwCode)
    {
        //  Enable touch panel
        case IOCTL_TOUCH_ENABLE_TOUCHPANEL:
            DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_TOUCH_ENABLE_TOUCHPANEL\r\n"));
            PDDTouchPanelEnable();
            dwResult = ERROR_SUCCESS;
            break;

            //  Disable touch panel
        case IOCTL_TOUCH_DISABLE_TOUCHPANEL:
            DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_TOUCH_DISABLE_TOUCHPANEL\r\n"));
            PDDTouchPanelDisable();
            dwResult = ERROR_SUCCESS;
            break;


            //  Get current sample rate
        case IOCTL_TOUCH_GET_SAMPLE_RATE:
            DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_TOUCH_GET_SAMPLE_RATE\r\n"));

            //  Check parameter validity
            if ((pBufOut == NULL) || (pdwActualOut == NULL) || (dwLenOut < sizeof(DWORD)))
            {
                dwResult = ERROR_INVALID_PARAMETER;
                break;
            }
            __try
            {
                *pdwActualOut = sizeof(DWORD);
                //  Get the sample rate
                *((DWORD*)pBufOut) = g_TouchDevice.nSampleRate;
                dwResult = ERROR_SUCCESS;
            }
            __except(EXCEPTION_EXECUTE_HANDLER)
            {
                DEBUGMSG(ZONE_ERROR, (L"TchPdd_Ioctl: Exception in IOCTL_TOUCH_SET_SAMPLE_RATE\r\n"));
            }

            break;

            //  Set the current sample rate
        case IOCTL_TOUCH_SET_SAMPLE_RATE:
            DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_TOUCH_SET_SAMPLE_RATE\r\n"));

            //  Check parameter validity
            if ((pBufIn == NULL) || (dwLenIn < sizeof(DWORD)))
            {
                dwResult = ERROR_INVALID_PARAMETER;
                break;
            }
            __try
            {
                //  Set the sample rate
                g_TouchDevice.nSampleRate = *((DWORD*)pBufIn);
                dwResult = ERROR_SUCCESS;
            }
            __except(EXCEPTION_EXECUTE_HANDLER)
            {
                DEBUGMSG(ZONE_ERROR, (L"TchPdd_Ioctl: Exception in IOCTL_TOUCH_GET_SAMPLE_RATE\r\n"));
            }

            break;

            //  Get touch properties
        case IOCTL_TOUCH_GET_TOUCH_PROPS:
            DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_TOUCH_GET_TOUCH_PROPS\r\n"));

			TCH_BSP_PROPS props;

            //  Check parameter validity
            if ((pBufOut == NULL) || (pdwActualOut == NULL) || (dwLenOut < sizeof(TCH_PROPS)))
            {
                dwResult = ERROR_INVALID_PARAMETER;
                break;
            }
            __try
            {
                *pdwActualOut = sizeof(TCH_PROPS);

                //  Fill out the touch driver properties
				((TCH_PROPS*)pBufOut)->minSampleRate          = SAMPLE_RATE_LOW_HZ;
				((TCH_PROPS*)pBufOut)->maxSampleRate          = SAMPLE_RATE_HIGH_HZ;
				((TCH_PROPS*)pBufOut)->minCalCount            = 5;
				((TCH_PROPS*)pBufOut)->touchType              = TOUCHTYPE_MULTITOUCH;
				((TCH_PROPS*)pBufOut)->xRangeMin              = SAMPLE_RANGE_MIN_VALUE;
				((TCH_PROPS*)pBufOut)->xRangeMax              = SAMPLE_RANGE_MAX_VALUE;
				((TCH_PROPS*)pBufOut)->yRangeMin              = SAMPLE_RANGE_MIN_VALUE;
				((TCH_PROPS*)pBufOut)->yRangeMax              = SAMPLE_RANGE_MAX_VALUE;
				((TCH_PROPS*)pBufOut)->calHoldSteadyTime      = CAL_HOLD_STEADY_TIME;
				((TCH_PROPS*)pBufOut)->calDeltaReset          = CAL_DELTA_RESET;
				if (g_TouchDevice.BspIfc.pfnGetProps((LPVOID)g_TouchDevice.hBspContext, &props))
				{
					((TCH_PROPS*)pBufOut)->maxSimultaneousSamples = props.dwMaxContact;
                	dwResult = ERROR_SUCCESS;
				}
				else
				{
					((TCH_PROPS*)pBufOut)->maxSimultaneousSamples = g_TouchDevice.dwMaxContact;
					dwResult = ERROR_NOT_READY;
				}
            }
            __except(EXCEPTION_EXECUTE_HANDLER)
            {
                DEBUGMSG(ZONE_ERROR, (L"TchPdd_Ioctl: Exception in IOCTL_TOUCH_GET_TOUCH_PROPS\r\n"));
            }
            break;

            //  Power management IOCTLs
        case IOCTL_POWER_CAPABILITIES:
            DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_POWER_CAPABILITIES\r\n"));

            //  Check parameter validity
            if ((pBufOut == NULL) || (pdwActualOut == NULL) || (dwLenOut < sizeof(POWER_CAPABILITIES)))
            {
                dwResult = ERROR_INVALID_PARAMETER;
                break;
            }
            __try
            {
                PPOWER_CAPABILITIES ppc = (PPOWER_CAPABILITIES) pBufOut;
                memset(ppc, 0, sizeof(*ppc));
                ppc->DeviceDx = DX_MASK(D0) | DX_MASK(D1) | DX_MASK(D2) | DX_MASK(D3) | DX_MASK(D4);

                *pdwActualOut = sizeof(POWER_CAPABILITIES);

                dwResult = ERROR_SUCCESS;
            }
            __except(EXCEPTION_EXECUTE_HANDLER)
            {
                DEBUGMSG(ZONE_ERROR, (L"TchPdd_Ioctl: Exception in IOCTL_POWER_CAPABILITIES\r\n"));
            }

            break;

        case IOCTL_POWER_GET:
            DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_POWER_GET\r\n"));

            //  Check parameter validity
            if ((pBufOut == NULL) || (pdwActualOut == NULL) || (dwLenOut < sizeof(CEDEVICE_POWER_STATE)))
            {
                dwResult = ERROR_INVALID_PARAMETER;
                break;
            }
            __try
            {
                *(PCEDEVICE_POWER_STATE) pBufOut = (CEDEVICE_POWER_STATE) g_TouchDevice.dwPowerState;

                *pdwActualOut = sizeof(CEDEVICE_POWER_STATE);

                dwResult = ERROR_SUCCESS;
            }
            __except(EXCEPTION_EXECUTE_HANDLER)
            {
                DEBUGMSG(ZONE_ERROR, (L"TchPdd_Ioctl: Exception in IOCTL_POWER_GET\r\n"));
            }
            break;

        case IOCTL_POWER_SET:
            DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_POWER_SET\r\n"));
            CEDEVICE_POWER_STATE dx;

            //  Check parameter validity
            if ((pBufOut == NULL) || (dwLenOut < sizeof(CEDEVICE_POWER_STATE)))
            {
                dwResult = ERROR_INVALID_PARAMETER;
                break;
            }
            __try
            {
                dx = *(CEDEVICE_POWER_STATE*)pBufOut;
            }
            __except(EXCEPTION_EXECUTE_HANDLER)
            {
                DEBUGMSG(ZONE_ERROR, (L"TchPdd_Ioctl: Exception in IOCTL_POWER_SET\r\n"));
                dwResult = ERROR_INVALID_PARAMETER;
                break;
            }

            if( VALID_DX(dx) )
            {
                DEBUGMSG(ZONE_FUNCTION, (L"TchPdd_Ioctl: IOCTL_POWER_SET = to D%u\r\n", dx));
                g_TouchDevice.dwPowerState = dx;

                if (pdwActualOut)
                    *pdwActualOut = sizeof(CEDEVICE_POWER_STATE);

                //  Enable touchscreen for D0-D2; otherwise disable
                switch( dx )
                {
                    case D0:
                    case D1:
                    case D2:
                        //  Enable touchscreen
                        PDDTouchPanelPowerHandler(FALSE);    //Enable Touch ADC
                        break;

                    case D3:
                    case D4:
                        //  Disable touchscreen
                        PDDTouchPanelPowerHandler(TRUE); //Disable Touch ADC
                        break;

                    default:
                        //  Ignore
                        break;
                }

                dwResult = ERROR_SUCCESS;
            }
            break;

        default:
            dwResult = ERROR_NOT_SUPPORTED;
            break;
    }

    if (dwResult != ERROR_SUCCESS)
    {
        SetLastError(dwResult);
        return FALSE;
    }
    return TRUE;
}


//==============================================================================
// Function Name: TchPdd_PowerUp
//
// Description: MDD passes xxx_PowerUp stream interface call to PDD.
//
// Arguments:   [IN] hPddContext. pddcontext returned in TchPdd_Init
//
// Ret Value:   None
//==============================================================================
void WINAPI TchPdd_PowerUp(
        DWORD hPddContext
        )
{
    DEBUGMSG(ZONE_FUNCTION, (TEXT("TchPdd_PowerUp+\r\n")));
    // Remove-W4: Warning C4100 workaround
    UNREFERENCED_PARAMETER(hPddContext);
    DEBUGMSG(ZONE_FUNCTION, (TEXT("TchPdd_PowerUp-\r\n")));
}

//==============================================================================
// Function Name: TchPdd_PowerDown
//
// Description: MDD passes xxx_PowerDown stream interface call to PDD.
//
// Arguments:   [IN] hPddContext. pddcontext returned in TchPdd_Init
//
// Ret Value:   None
//==============================================================================
void WINAPI TchPdd_PowerDown(
        DWORD hPddContext
        )
{
    DEBUGMSG(ZONE_FUNCTION, (TEXT("TchPdd_PowerDown+\r\n")));
    // Remove-W4: Warning C4100 workaround
    UNREFERENCED_PARAMETER(hPddContext);
    DEBUGMSG(ZONE_FUNCTION, (TEXT("TchPdd_PowerDown-\r\n")));
}


//==============================================================================
//Internal Functions
//==============================================================================

//==============================================================================
// Function Name: PDDCalibrationThread
//
// Description: This function is called from the PDD Init to calibrate the screen.
//              If the calibration data is already present in the registry,
//              this step is skipped, else a call is made into the GWES for calibration.
//
// Arguments:   None.
//
// Ret Value:   Success(1), faliure(0)
//==============================================================================
static HRESULT PDDCalibrationThread()
{
    HKEY hKey;
    DWORD dwType;
    LONG lResult;
    HANDLE hAPIs;

    DEBUGMSG(ZONE_FUNCTION, (TEXT("CalibrationThread+\r\n")));

    // try to open [HKLM\hardware\devicemap    ouch] key
    if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, RK_HARDWARE_DEVICEMAP_TOUCH, 0, KEY_ALL_ACCESS, &hKey))
    {
        DEBUGMSG(ZONE_CALIBRATE, (TEXT("CalibrationThread: calibration: Can't find [HKLM/%s]\r\n"), RK_HARDWARE_DEVICEMAP_TOUCH));
        return E_FAIL;
    }

    // check for calibration data (query the type of data only)
    lResult = RegQueryValueEx(hKey, RV_CALIBRATION_DATA, 0, &dwType, NULL, NULL);
    RegCloseKey(hKey);
    if (lResult == ERROR_SUCCESS)
    {
        // registry contains calibration data, return
        return S_OK;
    }

    hAPIs = OpenEvent(EVENT_ALL_ACCESS, FALSE, TEXT("SYSTEM/GweApiSetReady"));
    if (hAPIs)
    {
        WaitForSingleObject(hAPIs, INFINITE);
        CloseHandle(hAPIs);
    }

    // Perform calibration
    TouchCalibrate();

#if 1//def DEBUG
    TCHAR szCalibrationData[100];
    DWORD Size = sizeof(szCalibrationData);

    // try to open [HKLM\hardware\devicemap    ouch] key
    if (ERROR_SUCCESS != RegOpenKeyEx(HKEY_LOCAL_MACHINE, RK_HARDWARE_DEVICEMAP_TOUCH, 0, KEY_ALL_ACCESS, &hKey))
    {
        DEBUGMSG(ZONE_CALIBRATE, (TEXT("CalibrationThread: calibration: Can't find [HKLM/%s]\r\n"), RK_HARDWARE_DEVICEMAP_TOUCH));
        return E_FAIL;
    }

    // display new calibration data
    lResult = RegQueryValueEx(hKey, RV_CALIBRATION_DATA, 0, &dwType, NULL, NULL);
    if (lResult == ERROR_SUCCESS)
    {
        RegQueryValueEx(hKey, RV_CALIBRATION_DATA, 0, &dwType, (BYTE *) szCalibrationData, (DWORD *) &Size);
    }

    RETAILMSG(1, (TEXT("New Calibration data: %s\r\n"), szCalibrationData));
    RegCloseKey(hKey);
#endif

    DEBUGMSG(ZONE_FUNCTION, (TEXT("CalibrationThread-\r\n")));

    return S_OK;
}


//==============================================================================
// Function Name: PDDStartCalibrationThread
//
// Description: This function is creates the calibration thread with
//              PDDCalibrationThread as entry.
//
// Arguments:   None.
//
// Ret Value:   None
//==============================================================================
void PDDStartCalibrationThread()
{
    HANDLE hThread;

    hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)PDDCalibrationThread, NULL, 0, NULL);
    // We don't need the handle, close it here
    CloseHandle(hThread);
}

//==============================================================================
// Function: PDDTouchPanelPowerHandler
//
// This function indicates to the driver that the system is entering
// or leaving the suspend state.
//
// Parameters:
//      bOff
//          [in] TRUE indicates that the system is turning off. FALSE
//          indicates that the system is turning on.
//
// Returns:
//      None.
//==============================================================================
void PDDTouchPanelPowerHandler(BOOL boff)
{
    DEBUGMSG(ZONE_FUNCTION, (_T("PDDTouchPanelPowerHandler+\r\n")));

    if (g_TouchDevice.hTouchPanelEvent != NULL)
        if (g_TouchDevice.BspIfc.pfnPowerHandler)
        	g_TouchDevice.BspIfc.pfnPowerHandler((LPVOID)g_TouchDevice.hBspContext, boff);

    DEBUGMSG(ZONE_FUNCTION, (_T("PDDTouchPanelPowerHandler-\r\n")));

    return;
}

//==============================================================================
// Function Name: FreeIstData
//
// Description: This function cleanups the data which be used by the IST.
//
// Argument:
//      pArg
//          [in] The context data allocated in initialization.
// Returns:
//      If the operation succeeds eturns TRUE; otherwize returns FALSE.
//==============================================================================
static void FreeIstData(TOUCH_DEVICE *pTouch)
{
	if (pTouch) 
	{
		if (pTouch->pNowInputs)
		{
			LocalFree(pTouch->pNowInputs);
		}

		if (pTouch->pLastInputs)
		{
			LocalFree(pTouch->pLastInputs);
		}

		if (pTouch->pBspInputs)
		{
			LocalFree(pTouch->pBspInputs);
		}
	}
}

//==============================================================================
// Function Name: AllocateIstData
//
// Description: This function allocates the data which be used by the IST.
//
// Argument:
//      pArg
//          [in] The context data allocated in initialization.
// Returns:
//      If the operation succeeds eturns TRUE; otherwize returns FALSE.
//==============================================================================
static BOOL AllocateIstData(TOUCH_DEVICE *pTouch)
{
	const DWORD dwMaxContact = pTouch->dwMaxContact;
	BOOL fRet = FALSE;
	
	if (!pTouch) 
	{
        goto exit;
	}

    pTouch->pNowInputs = (PCETOUCHINPUT)LocalAlloc(LPTR, sizeof(CETOUCHINPUT) * dwMaxContact);
    if (pTouch->pNowInputs == NULL) {
        DEBUGMSG(ZONE_ERROR, (TEXT("%S: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        goto exit;
    }
	
    pTouch->pLastInputs = (PCETOUCHINPUT)LocalAlloc(LPTR, sizeof(CETOUCHINPUT) * dwMaxContact);
    if (pTouch->pLastInputs == NULL) {
        DEBUGMSG(ZONE_ERROR, (TEXT("%S: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        goto exit;
    }

    pTouch->pBspInputs = (TCH_BSP_INPUT_INFO *)LocalAlloc(LPTR, sizeof(TCH_BSP_INPUT_INFO) * dwMaxContact);
    if (pTouch->pBspInputs == NULL) {
        DEBUGMSG(ZONE_ERROR, (TEXT("%S: LocalAlloc error:%d\r\n"), __FUNCTION__, GetLastError()));
        goto exit;
    }

	for (DWORD i = 0; i < dwMaxContact; i++) 
	{
		pTouch->pLastInputs[i].dwFlags = TOUCHEVENTF_UP;
		pTouch->pLastInputs[i].dwID = (DWORD)-1;

		pTouch->pNowInputs[i].dwMask = TOUCHINPUTMASKF_CONTACTAREA;
		pTouch->pNowInputs[i].cxContact = 40;
		pTouch->pNowInputs[i].cyContact = 20;
		pTouch->pNowInputs[i].hSource = (HANDLE)1;
	}

	fRet = TRUE;
exit:
	if (!fRet)
	{
		FreeIstData(pTouch);
	}
	return fRet;
}

//==============================================================================
// Function Name: LookupInputFromId
//
// Description: This function looks up the Touch-Inputs which has contact ID which 
//              equals specified ID.
//
// Arguments:
//      pInput
//          [in] Touch-Inputs that will be look up.
//
//      Count
//          [in] Count of Touch-Inputs.
//
//      dwID
//          [in] Contact ID that will be compared.
//
// Returns:
//      the Touch-Input in according with dwID.
//==============================================================================
static PCETOUCHINPUT LookupInputFromId(PCETOUCHINPUT pInput, int Count, DWORD dwID)
{
	for (int i = 0; i < Count; i++) {
		if (pInput[i].dwID == dwID)
			return &pInput[i];
	}
	return NULL;
}

//==============================================================================
// Function Name: PDDTouchIST
//
// Description: This is the IST which waits on the touch event or Time out value.
//              Normally the IST waits on the touch event infinitely, but once the
//              pen down condition is recognized the time out interval is changed
//              to dwSamplingTimeOut.
//
// Argument:
//      pArg
//          [in] The context data allocated in initialization.
//
// Ret Value:   None
//==============================================================================
ULONG PDDTouchIST(PVOID   pArg)
{
	TOUCH_DEVICE *pTouch = (TOUCH_DEVICE *)pArg;
    PCETOUCHINPUT pNowInputs = NULL;
    PCETOUCHINPUT pLastInputs = NULL;
	TCH_BSP_INPUT_INFO * pBspInputs = NULL;
    INT count = 0, last_count = 0;

	static INT32 _x = 0, _y = 0;

    DEBUGMSG(ZONE_TOUCH_IST, (L"PDDTouchIST: IST thread started\r\n"));

    if (!pTouch->BspIfc.pfnGetSample)
	{
    	DEBUGMSG(ZONE_ERROR, (L"PDDTouchIST: BspGetSample function is not set\r\n"));
		goto exit; 
	}

	if (!AllocateIstData(pTouch))
	{
		goto exit; 
	}

    pNowInputs = pTouch->pNowInputs;
    pLastInputs = pTouch->pLastInputs;
	pBspInputs = pTouch->pBspInputs;

    //  Loop until told to stop
    while(!pTouch->bTerminateIST)
    {
        //  Wait for touch IRQ
        WaitForSingleObject(pTouch->hTouchPanelEvent, INFINITE);

        //  Check for terminate flag
        if (pTouch->bTerminateIST)
            break;
		// Get the inputs until the BSP layer is holding the inputs
    	while (pTouch->BspIfc.pfnGetSample((LPVOID)pTouch->hBspContext, pBspInputs, &count))
		{
			for (int i = 0; i < count; i++) 
			{
				PCETOUCHINPUT pInput = LookupInputFromId(pLastInputs, last_count, pBspInputs[i].dwID);
				pNowInputs[i].x = pBspInputs[i].x;
				pNowInputs[i].y = pBspInputs[i].y;
				pNowInputs[i].dwID = pBspInputs[i].dwID;
				pNowInputs[i].dwFlags = 0;
				if (pInput == NULL)
				{
					if (pBspInputs[i].Flags & TouchSampleDownFlag)
					{
						pNowInputs[i].dwFlags = TOUCHEVENTF_DOWN | TOUCHEVENTF_INRANGE;
					}
				} 
				else
				{
					if (pBspInputs[i].Flags & TouchSampleDownFlag)
					{
						if (pInput->dwFlags == TOUCHEVENTF_UP) 
						{
							pNowInputs[i].dwFlags = TOUCHEVENTF_DOWN | TOUCHEVENTF_INRANGE;
						}
						else
						{
							pNowInputs[i].dwFlags = TOUCHEVENTF_MOVE | TOUCHEVENTF_INRANGE;
						}
					}
					else if (pBspInputs[i].Flags == TouchSampleValidFlag)
					{
						pNowInputs[i].dwFlags = TOUCHEVENTF_UP;
					}
				}

			}
			if (count) 
			{
				memcpy(pLastInputs, pNowInputs, sizeof(CETOUCHINPUT) * count);
				s_pfnMddReportSampleSet(s_mddContext, count, pNowInputs);
				last_count = count;
			}
		}
    }

exit:
    if (pTouch->BspIfc.pfnInterruptDone)
	   	pTouch->BspIfc.pfnInterruptDone((LPVOID)pTouch->hBspContext);
    if (pTouch->BspIfc.pfnInterruptDisable)
	   	pTouch->BspIfc.pfnInterruptDisable((LPVOID)pTouch->hBspContext);

	FreeIstData(pTouch);

    CloseHandle(pTouch->hTouchPanelEvent);
    pTouch->hTouchPanelEvent = NULL;

    DEBUGMSG(ZONE_TOUCH_IST, (L"PDDTouchIST: IST thread ending\r\n"));
    return ERROR_SUCCESS;
}

//==============================================================================
// Function Name: PDDInitializeHardware
//
// Description: This routine configures the h/w
//
// Arguments:  None
//
// Ret Value:   TRUE - Success
//                   FAIL - Failure
//==============================================================================
BOOL PDDInitializeHardware(LPCTSTR pszActiveKey)
{
    BOOL   rc = FALSE;

    DEBUGMSG(ZONE_FUNCTION, (TEXT("PDDInitializeHardware+\r\n")));

    // Read parameters from registry
    if (GetDeviceRegistryParams(
                pszActiveKey,
                &g_TouchDevice,
                dimof(s_deviceRegParams),
                s_deviceRegParams) != ERROR_SUCCESS)
    {
        DEBUGMSG(ZONE_ERROR, (TEXT("ERROR: PDDInitializeHardware: Error reading from Registry.\r\n")));
        goto cleanup;
    }

    // Nothing to do here as of now, can do some BSP initialization here.
    rc = TRUE;


cleanup:

    DEBUGMSG(ZONE_FUNCTION, (TEXT("PDDInitializeHardware-\r\n")));
    if( rc == FALSE )
    {
        PDDDeinitializeHardware();
    }

    return rc;
}


//==============================================================================
// Function Name: PDDDeinitializeHardware
//
// Description: This routine Deinitializes the h/w
//
// Arguments:  None
//
// Ret Value:   None
//==============================================================================
VOID PDDDeinitializeHardware()
{
    DEBUGMSG(ZONE_FUNCTION, (TEXT("PDDDeinitializeHardware+\r\n")));

    // Nothing to do here as of now, can do some BSP deinitialization here.
    DEBUGMSG(ZONE_FUNCTION, (TEXT("PDDDeinitializeHardware-\r\n")));
}


//==============================================================================
// Function Name: PDDTouchPanelEnable
//
// Description: This routine creates the touch thread(if it is not already created)
//              initializes the interrupt and unmasks it.
//
// Arguments:  None
//
// Ret Value:   TRUE - Success
//==============================================================================
BOOL  PDDTouchPanelEnable()
{
    BOOL rc = TRUE;
    DEBUGMSG(ZONE_FUNCTION, (TEXT("PDDTouchPanelEnable+\r\n")));

    //  Check if already running
    if( s_hISTThread == NULL )
    {
        //  Clear terminate flag
        g_TouchDevice.bTerminateIST = FALSE;

        if (g_TouchDevice.BspIfc.pfnInterruptEnable)
			rc = g_TouchDevice.BspIfc.pfnInterruptEnable((LPVOID)g_TouchDevice.hBspContext);
		else
			rc = TRUE;

        g_TouchDevice.dwSamplingTimeOut = INFINITE;

        //  Create IST thread
        s_hISTThread = CreateThread( NULL, 0, PDDTouchIST, &g_TouchDevice, 0, NULL );
        if( s_hISTThread == NULL )
        {
            DEBUGMSG(ZONE_ERROR, (TEXT("PDDTouchPanelEnable: Failed to create IST thread\r\n")));
            return FALSE;
        }

        // set IST thread priority
        CeSetThreadPriority (s_hISTThread, g_TouchDevice.dwISTPriority);
    }

    DEBUGMSG(ZONE_FUNCTION, (TEXT("PDDTouchPanelEnable-\r\n")));
    return rc;
}

//==============================================================================
// Function Name: PDDTouchPanelDisable
//
// Description: This routine closes the IST and releases other resources.
//
// Arguments:  None
//
// Ret Value:   TRUE - Success
//==============================================================================
VOID  PDDTouchPanelDisable()
{
    DEBUGMSG(ZONE_FUNCTION, (TEXT("PDDTouchPanelDisable+\r\n")));

    //  Disable touch interrupt service thread
    if( g_TouchDevice.hTouchPanelEvent )
    {
        if (g_TouchDevice.BspIfc.pfnInterruptDone)
			g_TouchDevice.BspIfc.pfnInterruptDone((LPVOID)g_TouchDevice.hBspContext);
        if (g_TouchDevice.BspIfc.pfnInterruptDisable)
			g_TouchDevice.BspIfc.pfnInterruptDisable((LPVOID)g_TouchDevice.hBspContext);

        g_TouchDevice.bTerminateIST = TRUE;
        SetEvent( g_TouchDevice.hTouchPanelEvent );
        //Wait 3sec for the IST to exit and if the thread doesn't exit, forcefully terminate
        if (WaitForSingleObject(s_hISTThread, THREAD_EXIT_TIMEOUT) != WAIT_OBJECT_0)
        {
            // Do this only when the thread is not terminated properly.
            DEBUGCHK(FALSE);
            TerminateThread(s_hISTThread, (DWORD)-1);
        }
        if(s_hISTThread!=NULL)
        {
            CloseHandle(s_hISTThread);
            s_hISTThread = NULL;
        }
        if(g_TouchDevice.hTouchPanelEvent!=NULL)
        {
            CloseHandle(g_TouchDevice.hTouchPanelEvent);
            g_TouchDevice.hTouchPanelEvent=NULL;
        }
    }

    DEBUGMSG(ZONE_FUNCTION, (TEXT("PDDTouchPanelDisable-\r\n")));
}
